/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "misc.h"
#include "cfb.h"
#include "cfbmskbits.h"
#include "servermd.h"

/* Solid bresenham line */
/* NOTES
   e2 is used less often than e1, so it's not in a register
*/

#define XAXIS_R(e, e1, e3, nlw, bit, dst) \
{ \
    e += e1; \
    if (e >= 0) \
    { \
	(dst) += nlw; \
	e += e3; \
    } \
    if (bit == 31) \
    { \
	bit = -1; \
	(dst)++; \
    } \
    bit++; \
}

#define XAXIS_L(e, e1, e3, nlw, bit, dst) \
{ \
    e += e1; \
    if (e >= 0) \
    { \
	(dst) += nlw; \
	e += e3; \
    } \
    if (!bit) \
    { \
	bit = 32; \
	(dst)--; \
    } \
    bit--; \
}

#define YAXIS_R(e, e1, e3, nlw, bit, dst) \
{ \
    e += e1; \
    if (e >= 0) \
    { \
	if (bit == 31) \
	{ \
	    bit = -1; \
	    (dst)++; \
	} \
	bit++; \
	e += e3; \
    } \
    (dst) += nlw; \
}

#define YAXIS_L(e, e1, e3, nlw, bit, dst) \
{ \
    e += e1; \
    if (e >= 0) \
    { \
	if (!bit) \
	{ \
	    bit = 32; \
	    (dst)--; \
	} \
	bit--; \
	e += e3; \
    } \
    (dst) += nlw; \
}

cfbWinBresS(rop, pixel, planemask, depth, addrl, nlwidth, signdx, signdy, axis, x1, y1, e, e1, e2, len)
int rop;
unsigned long pixel;
unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
register int signdx;
int signdy;		/* signs of directions */
int axis;		/* major axis (Y_AXIS or X_AXIS) */
int x1, y1;		/* initial point */
register int e;		/* error accumulator */
register int e1;	/* bresenham increments */
int e2;
register int len;	/* length of line */
{
    register int e3 = e2-e1;
    register int bit;

    if (len == 0)
    	return;

    sx_check_transparent(rop, ALL_ONE, pixel, planemask);
    sx_plane_select(planemask);

    /* point to longword containing first point */
    addrl = (addrl + (y1 * nlwidth) + (x1 >> PWSH));
    if (signdy < 0)
	    nlwidth = -nlwidth;
    e = e-e1;			/* to make looping easier */

    bit = x1 & PIM;

    if (axis == X_AXIS)
    {
	if (signdx > 0)
	{
	    do {
		*addrl = cfbmask[bit];
		XAXIS_R(e, e1, e3, nlwidth, bit, addrl);
	    }
	    while(--len);
	}
	else
	{
	    do {
		*addrl = cfbmask[bit];
		XAXIS_L(e, e1, e3, nlwidth, bit, addrl);
	    }
	    while(--len);
	}
    } /* if X_AXIS */
    else
    {
	if (signdx > 0)
	{
	    do {
		*addrl = cfbmask[bit];
		YAXIS_R(e, e1, e3, nlwidth, bit, addrl);
	    }
	    while(--len);
	}
	else
	{
	    do {
		*addrl = cfbmask[bit];
		YAXIS_L(e, e1, e3, nlwidth, bit, addrl);
	    }
	    while(--len);
	}
    } /* else Y_AXIS */
}

#define WriteBitData(ROP) { \
    Poffset = cfbploff; \
    for (plane = 0; plane < depth; plane++) \
    { \
	if (cfbplmk[plane]) \
	{ \
	    pdst = addrl + *Poffset; \
	    *pdst = (unsigned int) \
		((ROP(cfbfgmk[plane],*pdst) & cfbmask[bit]) | \
		(*pdst & cfbrmask[bit])); \
	} \
	Poffset++; \
    } \
}


#define BRS(ROP) \
    if (axis == X_AXIS) \
    { \
	if (signdx > 0) \
	{ \
	    do { \
		WriteBitData(ROP) \
		XAXIS_R(e, e1, e3, nlwidth, bit, addrl); \
	    } \
	    while(--len); \
	} \
	else \
	{ \
	    do { \
		WriteBitData(ROP) \
		XAXIS_L(e, e1, e3, nlwidth, bit, addrl); \
	    } \
	    while(--len); \
	} \
    } \
    else \
    { \
	if (signdx > 0) \
	{ \
	    do { \
		WriteBitData(ROP) \
		YAXIS_R(e, e1, e3, nlwidth, bit, addrl); \
	    } \
	    while(--len); \
	} \
	else \
	{ \
	    do { \
		WriteBitData(ROP) \
		YAXIS_L(e, e1, e3, nlwidth, bit, addrl); \
	    } \
	    while(--len); \
	} \
    }

cfbPixBresS(rop, pixel, planemask, depth, addrl, nlwidth, signdx, signdy, axis, x1, y1, e, e1, e2, len)
int rop;
unsigned long pixel;
unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
register int signdx;
int signdy;		/* signs of directions */
int axis;		/* major axis (Y_AXIS or X_AXIS) */
int x1, y1;		/* initial point */
register int e;		/* error accumulator */
register int e1;	/* bresenham increments */
int e2;
register int len;	/* length of line */
{
    register int e3 = e2-e1;
    register long bit;
    register int *Poffset;
    register int *pdst;
    int plane;

    if (len == 0)
    	return;

    /* point to longword containing first point */
    addrl = (addrl + (y1 * nlwidth) + (x1 >> PWSH));
    if (signdy < 0)
	    nlwidth = -nlwidth;
    e = e-e1;			/* to make looping easier */

    bit = x1 & PIM;

    switch(rop)
    {
	case GXcopy:
	    BRS(fnCOPY);
	    break;
	case GXxor:
	    BRS(fnXOR);
	    break;
	case GXand:
	    BRS(fnAND);
	    break;
	case GXor:
	    BRS(fnOR);
	    break;
	case GXset:
	    BRS(fnSET);
	    break;
	case GXclear:
	    BRS(fnCLEAR);
	    break;
	case GXandReverse:
	    BRS(fnANDREVERSE);
	    break;
	case GXandInverted:
	    BRS(fnANDINVERTED);
	    break;
	case GXnor:
	    BRS(fnNOR);
	    break;
	case GXequiv:
	    BRS(fnEQUIV);
	    break;
	case GXinvert:
	    BRS(fnINVERT);
	    break;
	case GXorReverse:
	    BRS(fnORREVERSE);
	    break;
	case GXcopyInverted:
	    BRS(fnCOPYINVERTED);
	    break;
	case GXorInverted:
	    BRS(fnORINVERTED);
	    break;
	case GXnand:
	    BRS(fnNAND);
	    break;
	default:
	    break;
    }
}
