/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "misc.h"
#include "cfb.h"
#include "cfbmskbits.h"

/* Dashed bresenham line */

#define StepDashWin\
    if (!--dashRemaining) { \
	if (++ dashIndex == numInDashList) \
	    dashIndex = 0; \
	dashRemaining = pDash[dashIndex]; \
	if (isDoubleDash) { \
	    pixel = fg; \
	    if (dashIndex & 1) \
		pixel = bg; \
	    if (oldpixel != pixel) { \
		oldpixel = pixel; \
		sx_check_transparent(rop, ALL_ONE, pixel, planemask); \
		sx_plane_select(planemask); \
	    } \
	} \
	else \
	    dontdraw = (dashIndex & 1); \
    }

#define StepDashPix\
    if (!--dashRemaining) { \
	if (++ dashIndex == numInDashList) \
	    dashIndex = 0; \
	dashRemaining = pDash[dashIndex]; \
	if (isDoubleDash) { \
	    pixel = fg; \
	    if (dashIndex & 1) \
		pixel = bg; \
	    if (oldpixel != pixel) { \
		oldpixel = pixel; \
		if (pixel == fg) { \
		    Pixel = cfbfgmk; \
		} \
		else { \
		    Pixel = cfbbgmk; \
		} \
	    } \
	} \
	else \
	    dontdraw = (dashIndex & 1); \
    }

#define XAXIS_R(e, e1, e3, nlw, bit, dst)\
{ \
    e += e1; \
    if (e >= 0) \
    { \
	(dst) += nlw; \
	e += e3; \
    } \
    if (bit == 31) \
    { \
	bit = -1; \
	(dst)++; \
    } \
    bit++; \
}

#define XAXIS_L(e, e1, e3, nlw, bit, dst)\
{ \
    e += e1; \
    if (e >= 0) \
    { \
	(dst) += nlw; \
	e += e3; \
    } \
    if (!bit) \
    { \
	bit = 32; \
	(dst)--; \
    } \
    bit--; \
}

#define YAXIS_R(e, e1, e3, nlw, bit, dst)\
{ \
    e += e1; \
    if (e >= 0) \
    { \
	if (bit == 31) \
	{ \
	    bit = -1; \
	    (dst)++; \
	} \
	bit++; \
	e += e3; \
    } \
    (dst) += nlw; \
}

#define YAXIS_L(e, e1, e3, nlw, bit, dst)\
{ \
    e += e1; \
    if (e >= 0) \
    { \
	if (!bit) \
	{ \
	    bit = 32; \
	    (dst)--; \
	} \
	bit--; \
	e += e3; \
    } \
    (dst) += nlw; \
}

cfbWinBresD(rop, fg, bg, planemask, depth,
	 pdashIndex, pDash, numInDashList, pdashOffset, isDoubleDash,
	 addrl, nlwidth,
	 signdx, signdy, axis, x1, y1, e, e1, e2, len)
int rop;
unsigned long fg, bg;
unsigned long planemask;
int depth;
int *pdashIndex;	/* current dash */
unsigned char *pDash;	/* dash list */
int numInDashList;	/* total length of dash list */
int *pdashOffset;	/* offset into current dash */
int isDoubleDash;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
int signdx, signdy;	/* signs of directions */
int axis;		/* major axis (Y_AXIS or X_AXIS) */
int x1, y1;		/* initial point */
register int e;		/* error accumulator */
register int e1;	/* bresenham increments */
int e2;
register int len;	/* length of line */
{
    register int yinc;	/* increment to next scanline, in bytes */
    register int e3 = e2-e1;
    int dashIndex;
    int dashOffset;
    int dashRemaining;
    unsigned long pixel;
    int dontdraw;
    register long bit;
    int oldpixel;

    if (len == 0)
	return;

    dashOffset = *pdashOffset;
    dashIndex = *pdashIndex;
    dashRemaining = pDash[dashIndex] - dashOffset;
    pixel = fg;
    if (isDoubleDash)
    {
	if (dashIndex & 1)
	    pixel = bg;
	dontdraw = 0;
    }
    else
	dontdraw = (dashIndex & 1);

    oldpixel = pixel;
    sx_check_transparent(rop, ALL_ONE, pixel, planemask);
    sx_plane_select(planemask);

    /* point to longword containing first point */
    addrl = (addrl + (y1 * nlwidth) + (x1 >> PWSH));
    yinc = signdy * nlwidth;
    e = e-e1;			/* to make looping easier */

    bit = x1 & PIM;

    if (axis == X_AXIS)
    {
	if (signdx > 0)
	{
	    do {
		if (!dontdraw)
		{
		    *addrl = cfbmask[bit];
		}
		XAXIS_R(e, e1, e3, yinc, bit, addrl);
		StepDashWin
	    }
	    while (--len);
	}
	else
	{
	    do {
		if (!dontdraw)
		{
		    *addrl = cfbmask[bit];
		}
		XAXIS_L(e, e1, e3, yinc, bit, addrl);
		StepDashWin
	    }
	    while (--len);
	}
    } /* if X_AXIS */
    else
    {
	if (signdx > 0)
	{
	    do {
		if (!dontdraw)
		{
		    *addrl = cfbmask[bit];
		}
		YAXIS_R(e, e1, e3, yinc, bit, addrl);
		StepDashWin
	    }
	    while(--len);
	}
	else
	{
	    do {
		if (!dontdraw)
		{
		    *addrl = cfbmask[bit];
		}
		YAXIS_L(e, e1, e3, yinc, bit, addrl);
		StepDashWin
	    }
	    while(--len);
	}
    } /* else Y_AXIS */
    *pdashIndex = dashIndex;
    *pdashOffset = pDash[dashIndex] - dashRemaining;
}

#define WriteBitData(ROP) { \
    Poffset = cfbploff; \
    for (plane = 0; plane < depth; plane++) \
    { \
	if (cfbplmk[plane]) \
	{ \
	     pdst = addrl + *Poffset; \
	     *pdst = (unsigned int) \
		((ROP(Pixel[plane],*pdst) & cfbmask[bit]) | \
		(*pdst & cfbrmask[bit])); \
	} \
	Poffset++; \
    }  \
}

#define BRSD(ROP) \
    if (axis == X_AXIS) \
    { \
	if (signdx > 0) \
	{ \
	    do { \
		if (!dontdraw)  \
		{  \
		    WriteBitData(ROP) \
		} \
		XAXIS_R(e, e1, e3, yinc, bit, addrl); \
		StepDashPix \
	    } \
	    while(--len); \
	} \
	else \
	{ \
	    do { \
		if (!dontdraw)  \
		{  \
		    WriteBitData(ROP) \
		} \
		XAXIS_L(e, e1, e3, yinc, bit, addrl); \
		StepDashPix \
	    } \
	    while(--len); \
	} \
    } \
    else \
    { \
	if (signdx > 0) \
	{ \
	    do { \
		if (!dontdraw)  \
		{  \
		    WriteBitData(ROP) \
		}  \
		YAXIS_R(e, e1, e3, yinc, bit, addrl); \
		StepDashPix \
	    } \
	    while(--len); \
	} \
	else \
	{ \
	    do { \
		if (!dontdraw)  \
		{  \
		    WriteBitData(ROP) \
		} \
		YAXIS_L(e, e1, e3, yinc, bit, addrl); \
		StepDashPix \
	    } \
	    while(--len); \
	} \
    }

cfbPixBresD(rop, fg, bg, planemask, depth,
	 pdashIndex, pDash, numInDashList,
	 pdashOffset, isDoubleDash, addrl, nlwidth,
	 signdx, signdy, axis, x1, y1, e, e1, e2, len)
int rop;
unsigned long fg, bg;
unsigned long planemask;
int depth;
int *pdashIndex;	/* current dash */
unsigned char *pDash;	/* dash list */
int numInDashList;	/* total length of dash list */
int *pdashOffset;	/* offset into current dash */
int isDoubleDash;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
int signdx, signdy;	/* signs of directions */
int axis;		/* major axis (Y_AXIS or X_AXIS) */
int x1, y1;		/* initial point */
register int e;		/* error accumulator */
register int e1;	/* bresenham increments */
int e2;
register int len;	/* length of line */
{
    register int yinc;	/* increment to next scanline, in bytes */
    register int e3 = e2-e1;
    int dashIndex;
    int dashOffset;
    int dashRemaining;
    unsigned long pixel;
    int dontdraw;
    register long bit;
    int oldpixel, plane;
    register unsigned long *Pixel;
    register int *Poffset;
    register int *pdst;

    if (len == 0)
	return;

    dashOffset = *pdashOffset;
    dashIndex = *pdashIndex;
    dashRemaining = pDash[dashIndex] - dashOffset;
    pixel = fg;
    if (isDoubleDash)
    {
	if (dashIndex & 1)
	    pixel = bg;
	dontdraw = 0;
    }
    else
	dontdraw = (dashIndex & 1);

    oldpixel = pixel;
    if (pixel == fg)
    {
	Pixel = cfbfgmk;
    }
    else
    {
	Pixel = cfbbgmk;
    }

    /* point to longword containing first point */
    addrl = (addrl + (y1 * nlwidth) + (x1 >> PWSH));
    yinc = signdy * nlwidth;
    e = e-e1;			/* to make looping easier */

    bit = x1 & PIM;

    switch(rop)
    {
	case GXcopy:
	    BRSD(fnCOPY);
	    break;
	case GXxor:
	    BRSD(fnXOR);
	    break;
	case GXand:
	    BRSD(fnAND);
	    break;
	case GXor:
	    BRSD(fnOR);
	    break;
	case GXset:
	    BRSD(fnSET);
	    break;
	case GXclear:
	    BRSD(fnCLEAR);
	    break;
	case GXandReverse:
	    BRSD(fnANDREVERSE);
	    break;
	case GXandInverted:
	    BRSD(fnANDINVERTED);
	    break;
	case GXnor:
	    BRSD(fnNOR);
	    break;
	case GXequiv:
	    BRSD(fnEQUIV);
	    break;
	case GXinvert:
	    BRSD(fnINVERT);
	    break;
	case GXorReverse:
	    BRSD(fnORREVERSE);
	    break;
	case GXcopyInverted:
	    BRSD(fnCOPYINVERTED);
	    break;
	case GXorInverted:
	    BRSD(fnORINVERTED);
	    break;
	case GXnand:
	    BRSD(fnNAND);
	    break;
	default:
	    break;
    }
    *pdashIndex = dashIndex;
    *pdashOffset = pDash[dashIndex] - dashRemaining;
}
