/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

/*
 * Fast bitblt macros for certain hardware.  If your machine has an addressing
 * mode of small constant + register, you'll probably want this magic specific
 * code.  It's 25% faster for the R2000.  I haven't studied the Sparc
 * instruction set, but I suspect it also has this addressing mode.  Also,
 * unrolling the loop by 32 is possibly excessive for mfb. The number of times
 * the loop is actually looped through is pretty small.
 */

/*
 * WARNING:  These macros make *a lot* of assumptions about
 * the environment they are invoked in.  Plenty of implicit
 * arguments, lots of side effects.  Don't use them casually.
 */

#define SwitchOdd(n) case n: BodyOdd(n)
#define SwitchEven(n) case n: BodyEven(n)

/* to allow mfb and cfb to share code... */
#ifndef BitRight
#define BitRight(a,b) SCRRIGHT(a,b)
#define BitLeft(a,b) SCRLEFT(a,b)
#endif

#ifdef LARGE_INSTRUCTION_CACHE
#define UNROLL 8
#define PackedLoop \
    switch (nl & (UNROLL-1)) { \
    SwitchOdd( 7) SwitchEven( 6) SwitchOdd( 5) SwitchEven( 4) \
    SwitchOdd( 3) SwitchEven( 2) SwitchOdd( 1) \
    } \
    while ((nl -= UNROLL) >= 0) { \
	LoopReset \
	BodyEven( 8) \
    	BodyOdd( 7) BodyEven( 6) BodyOdd( 5) BodyEven( 4) \
    	BodyOdd( 3) BodyEven( 2) BodyOdd( 1) \
    }
#else
#define UNROLL 4
#define PackedLoop \
    switch (nl & (UNROLL-1)) { \
    SwitchOdd( 3) SwitchEven( 2) SwitchOdd( 1) \
    } \
    while ((nl -= UNROLL) >= 0) { \
	LoopReset \
    	BodyEven( 4) \
    	BodyOdd( 3) BodyEven( 2) BodyOdd( 1) \
    }
#endif

#if	1
#define DuffL(counter,label,body) \
    switch (counter & 3) { \
    label: \
        body \
    case 3: \
	body \
    case 2: \
	body \
    case 1: \
	body \
    case 0: \
	if ((counter -= 4) >= 0) \
	    goto label; \
    }
#else
#define DuffL(counter,label,body) \
    switch (counter & 7) { \
    label: \
	body \
    case 7: \
        body \
    case 6: \
	body \
    case 5: \
	body \
    case 4: \
	body \
    case 3: \
	body \
    case 2: \
	body \
    case 1: \
	body \
    case 0: \
	if ((counter -= 8) >= 0) \
	    goto label; \
    }
#endif


/* addded OMRON */

/* bitblt function jump table */
typedef struct {
	void (* fbmem)();
	void (* mem)();
} FUNCTBL;

#ifdef	LSB_IS_P0
#define checkRevSrc( pbase, srcoff, depth )
#define	planeLoop(pmask)	while(pmask) 
#define	initPmask(pmask,depth)	pmask <<= 32 - (depth)
#define bitmapPmask()		0x80000000
#define	checkPmask(pmask)	(pmask) & 0x80000000
#define	updatePmask(pmask) 	(pmask) <<= 1
#else
#define checkRevSrc( pbase, srcoff, depth ) \
{ \
	if(cfbPlaneRevFlag) \
	{ \
		pbase += (srcoff) * (depth - 1); \
		(srcoff) = - (srcoff); \
	} \
}
#define	planeLoop(pmask)	while(pmask) 
#define	initPmask(pmask,depth)	pmask &= (1 << depth) -1
#define bitmapPmask()		1
#define	checkPmask(pmask)	(pmask) & 1
#define	updatePmask(pmask) 	(pmask) >>= 1
#endif


#define	slow_putbits_port1( tmpbits, x, tmpmask, pdst)	\
{ \
	*(pdst) = (*(pdst) & ~tmpmask) | \
	    (SCRRIGHT((unsigned) tmpbits, x) & tmpmask); \
}

#define	slow_rop_putbits_port1( tmpbits, x, tmpmask, pdst, FUNC)	\
{ \
	*(pdst) = (*(pdst) & ~tmpmask) | \
	    ( (FUNC (SCRRIGHT((unsigned) tmpbits, x), *(pdst))) & tmpmask); \
}


#define	setWindowBaseAndWidth( pDraw, pBase, width, PlaneOffset) \
{ \
    pBase = (unsigned int *) \
	(((PixmapPtr)(pDraw->pScreen->devPrivate))->devPrivate.ptr); \
    width = (int) \
	   ((PixmapPtr)(pDraw->pScreen->devPrivate))->devKind >> 2; \
    PlaneOffset = getFbPlaneLongSize(); \
}


#define	setPixmapBaseAndWidth( pDraw, pBase, width, PlaneOffset) \
{ \
    pBase = (unsigned int *)(((PixmapPtr)pDraw)->devPrivate.ptr); \
    width = (int)(((PixmapPtr)pDraw)->devKind) >> 2; \
    PlaneOffset =  width * pDraw->height; \
}


#define	setBaseAndWidth( pDraw, pBase, width, PlaneOffset) \
{ \
    if (pDraw->type == DRAWABLE_WINDOW) \
    { \
	setWindowBaseAndWidth( pDraw, pBase, width, PlaneOffset) \
    } \
    else \
    { \
	setPixmapBaseAndWidth( pDraw, pBase, width, PlaneOffset) \
    } \
}


#define	checkAndChangeRectBox( careful, pptSrc, pbox, nbox, pboxNew1, pptNew1, \
			       pboxNew2, pptNew, ydir, xdir) \
{   BoxPtr pboxTmp, pboxNext, pboxBase; \
    DDXPointPtr pptTmp; \
    pboxNew1 = NULL; \
    pptNew1 = NULL; \
    pboxNew2 = NULL; \
    pptNew2 = NULL; \
    if (careful && (pptSrc->y < pbox->y1)) \
    {   /* walk source botttom to top */ \
	ydir = -1; \
	widthSrc = -widthSrc; \
	widthDst = -widthDst; \
	if (nbox > 1) \
	{   /* keep ordering in each band, reverse order of bands */ \
	    pboxNew1 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox); \
	    if(!pboxNew1) \
		return; \
	    pptNew1 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox); \
	    if(!pptNew1) \
	    { \
	        DEALLOCATE_LOCAL(pboxNew1); \
	        return; \
	    } \
	    pboxBase = pboxNext = pbox+nbox-1; \
	    while (pboxBase >= pbox) \
	    { \
	        while ((pboxNext >= pbox) &&  \
		       (pboxBase->y1 == pboxNext->y1)) \
		    pboxNext--; \
	        pboxTmp = pboxNext+1; \
	        pptTmp = pptSrc + (pboxTmp - pbox); \
	        while (pboxTmp <= pboxBase) \
	        { \
		    *pboxNew1++ = *pboxTmp++; \
		    *pptNew1++ = *pptTmp++; \
	        } \
	        pboxBase = pboxNext; \
	    } \
	    pboxNew1 -= nbox; \
	    pbox = pboxNew1; \
	    pptNew1 -= nbox; \
	    pptSrc = pptNew1; \
        } \
    } \
    else \
    { \
	/* walk source top to bottom */ \
	ydir = 1; \
    } \
    if (careful && (pptSrc->x < pbox->x1)) \
    { /* walk source right to left */ \
        xdir = -1; \
	if (nbox > 1) \
	{ /* reverse order of rects in each band */ \
	    pboxNew2 = (BoxPtr)ALLOCATE_LOCAL(sizeof(BoxRec) * nbox); \
	    pptNew2 = (DDXPointPtr)ALLOCATE_LOCAL(sizeof(DDXPointRec) * nbox); \
	    if(!pboxNew2 || !pptNew2) \
	    { \
		if (pptNew2) DEALLOCATE_LOCAL(pptNew2); \
		if (pboxNew2) DEALLOCATE_LOCAL(pboxNew2); \
		if (pboxNew1) \
		{ \
		    DEALLOCATE_LOCAL(pptNew1); \
		    DEALLOCATE_LOCAL(pboxNew1); \
		} \
	        return; \
	    } \
	    pboxBase = pboxNext = pbox; \
	    while (pboxBase < pbox+nbox) \
	    { \
	        while ((pboxNext < pbox+nbox) && \
		       (pboxNext->y1 == pboxBase->y1)) \
		    pboxNext++; \
	        pboxTmp = pboxNext; \
	        pptTmp = pptSrc + (pboxTmp - pbox); \
	        while (pboxTmp != pboxBase) \
	        { \
		    *pboxNew2++ = *--pboxTmp; \
		    *pptNew2++ = *--pptTmp; \
	        } \
	        pboxBase = pboxNext; \
	    } \
	    pboxNew2 -= nbox; \
	    pbox = pboxNew2; \
	    pptNew2 -= nbox; \
	    pptSrc = pptNew2; \
	} \
    } \
    else \
    { /* walk source left to right */ \
        xdir = 1; \
    } \
}



#define	FB_bitblt_fill_area(nbox, pbox, pdstBase, widthDst, alu, pmask, \
			    LABEL) \
{   unsigned int *pdstLine; /* pointers to line with current src and dst */ \
                            /* following used for looping through a line */ \
    int w, h; \
    register unsigned int *pdst; \
    register int xoffDst;   \
 \
    sx_check_transparent(alu, ~0, pmask, pmask); \
    sx_plane_select(pmask); \
 \
    while(nbox--) \
    { \
	w = pbox->x2 - pbox->x1; \
	h = pbox->y2 - pbox->y1; \
	pdstLine = pdstBase + (pbox->y1 * widthDst); \
	xoffDst = pbox->x1 & PIM; \
 \
	pdstLine += (pbox->x1 >> PWSH); \
	if ((xoffDst) + w <= PPW) /* width < PPW */ \
	{   register int tmpmask; \
	    maskpartialbits((xoffDst), (w), tmpmask); \
	    pdst = pdstLine; \
	    while (h--) \
	    { \
		*pdst = tmpmask; \
		pdst += widthDst; \
	    } \
	} \
	else	/* width > PPW */ \
	{   unsigned int startmask, endmask; \
	    int nlMiddle;	/* whole longwords in dst */ \
    	    register int nl;	/* temp copy of nlMiddle */ \
 \
    	    maskbits(pbox->x1, w, startmask, endmask, nlMiddle); \
 \
	    while (h--) \
	    {    \
	    	pdst = pdstLine; \
		if (startmask) \
		{ \
		    *pdst = startmask; \
		    pdst++; \
		} \
	    	nl = nlMiddle; \
	    	DuffL(nl, LABEL, *pdst++ = ~0;) \
		if (endmask) \
		{ \
	    	    *pdst = endmask; \
		} \
	    	pdstLine += widthDst; \
	    }  /* h */ \
	} /* width */ \
	pbox++; \
    } /* nbox */ \
}

extern int cfbPlaneRevFlag;
