/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"

#include "gc.h"
#include "window.h"
#include "pixmap.h"
#include "region.h"

#include "cfb.h"
#include "cfbmskbits.h"

/* horizontal solid line
   abs(len) > 1
*/
cfbWinHorzS(rop, pixel, planemask, depth, addrl, nlwidth, x1, y1, len)
register int rop;
register unsigned long pixel;
register unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
int x1;			/* initial point */ 
int y1;
int len;		/* length of line */
{
    register int nlmiddle;
    register int startmask;
    register int endmask;

    sx_check_transparent(rop, ALL_ONE, pixel, planemask);
    sx_plane_select(planemask);

    /* force the line to go left to right
       but don't draw the last point
    */
    if (len < 0)
    {
	x1 += len;
	x1 += 1;
	len = -len;
    }

    addrl = addrl + (y1 * nlwidth) + (x1 >> PWSH);

    /* all bits inside same longword */
    if ( ((x1 & PIM) + len) < PPW)
    {
	maskpartialbits(x1, len, startmask);
	*addrl = startmask;
    }
    else
    {
	maskbits(x1, len, startmask, endmask, nlmiddle);
	if (startmask)
	{
	    *addrl++ = startmask;
	}
	Duff(nlmiddle, *addrl++ = ALL_ONE;);
	if (endmask)
	{
   	    *addrl = endmask;
	}
    }
}

#define HORZ(ROP) \
    for (plane = 0; plane < depth; plane++) \
    { \
	if (cfbplmk[plane]) \
	{ \
	    *addrl = ((ROP(cfbfgmk[plane], *addrl) & startmask) | \
			(*addrl & ~startmask)); \
	} \
	addrl += Poffset; \
    }

#define HORZODD(ROP) \
    for (plane = 0; plane < depth; plane++) \
    { \
	addrl = Addrl + cfbploff[plane]; \
	if (cfbplmk[plane]) \
	{ \
	    if (startmask) \
	    { \
		*addrl++ = ((ROP(cfbfgmk[plane], *addrl) & startmask ) |  \
		    (*addrl & ~startmask)); \
	    } \
	    nlwidth = nlmiddle; \
	    while (nlwidth--) \
	    { \
		*addrl++ = ROP(cfbfgmk[plane], *addrl); \
	    } \
	    if (endmask) \
	    { \
		*addrl = ((ROP(cfbfgmk[plane], *addrl) & endmask ) | \
		    (*addrl & ~endmask)); \
	    } \
	} \
    }

cfbPixHorzS(rop, pixel, planemask, depth, addrl, nlwidth, x1, y1, len)
int rop;
unsigned long pixel;
unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
int nlwidth;		/* width in longwords of bitmap */
int x1;			/* initial point */ 
int y1;
int len;		/* length of line */
{
    register int nlmiddle;
    register int startmask;
    register int endmask;
    register int plane;

    /* force the line to go left to right
       but don't draw the last point
    */
    if (len < 0)
    {
	x1 += len;
	x1 += 1;
	len = -len;
    }

    addrl = addrl + (y1 * nlwidth) + (x1 >> PWSH);

    /* all bits inside same longword */
    if ( ((x1 & PIM) + len) < PPW)
    {
	register int Poffset = cfbploff[1];

	maskpartialbits(x1, len, startmask);
	switch(rop)
	{
	    case GXcopy:
		HORZ(fnCOPY);
		break;
	    case GXxor:
		HORZ(fnXOR);
		break;
	    case GXand:
		HORZ(fnAND);
		break;
	    case GXor:
		HORZ(fnOR);
		break;
	    case GXset:
		HORZ(fnSET);
		break;
	    case GXclear:
		HORZ(fnCLEAR);
		break;
	    case GXandReverse:
		HORZ(fnANDREVERSE);
		break;
	    case GXandInverted:
		HORZ(fnANDINVERTED);
		break;
	    case GXnor:
		HORZ(fnNOR);
		break;
	    case GXequiv:
		HORZ(fnEQUIV);
		break;
	    case GXinvert:
		HORZ(fnINVERT);
		break;
	    case GXorReverse:
		HORZ(fnORREVERSE);
		break;
	    case GXcopyInverted:
		HORZ(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		HORZ(fnORINVERTED);
		break;
	    case GXnand:
		HORZ(fnNAND);
		break;
	    default:
		break;
	}
    }
    else
    {
	register int *Addrl;

	maskbits(x1, len, startmask, endmask, nlmiddle);

	Addrl = addrl;
	switch(rop)
	{
	    case GXcopy:
		HORZODD(fnCOPY);
		break;
	    case GXxor:
		HORZODD(fnXOR);
		break;
	    case GXand:
		HORZODD(fnAND);
		break;
	    case GXor:
		HORZODD(fnOR);
		break;
	    case GXset:
		HORZODD(fnSET);
		break;
	    case GXclear:
		HORZODD(fnCLEAR);
		break;
	    case GXandReverse:
		HORZODD(fnANDREVERSE);
		break;
	    case GXandInverted:
		HORZODD(fnANDINVERTED);
		break;
	    case GXnor:
		HORZODD(fnNOR);
		break;
	    case GXequiv:
		HORZODD(fnEQUIV);
		break;
	    case GXinvert:
		HORZODD(fnINVERT);
		break;
	    case GXorReverse:
		HORZODD(fnORREVERSE);
		break;
	    case GXcopyInverted:
		HORZODD(fnCOPYINVERTED);
		break;
	    case GXorInverted:
		HORZODD(fnORINVERTED);
		break;
	    case GXnand:
		HORZODD(fnNAND);
		break;
	    default:
		break;
	}
    }
}

/* vertical solid line
   this uses do loops because pcc (Ultrix 1.2, bsd 4.2) generates
   better code.  sigh.  we know that len will never be 0 or 1, so
   it's OK to use it.
*/

cfbWinVertS(rop, pixel, planemask, depth, addrl, nlwidth, x1, y1, len)
int rop;
register unsigned long pixel;
unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
register int nlwidth;	/* width in longwords of bitmap */
int x1, y1;		/* initial point */
register int len;	/* length of line */
{
    register unsigned long bitmask = cfbmask[x1 & PIM];

    sx_check_operation(rop, bitmask, pixel, planemask);
    sx_plane_select(planemask);

    addrl = addrl + (y1 * nlwidth) + (x1 >> PWSH);

    if (len < 0)
    {
	nlwidth = -nlwidth;
	len = -len;
    }

    do
    {
	*addrl = bitmask;
	addrl += nlwidth;
    }
    while (--len);
}

#define VERT(ROP) \
    do \
    { \
	for (plane = 0; plane < depth; plane++) \
	{ \
	    if (cfbplmk[plane]) \
	    { \
		*addrl = ((ROP(cfbfgmk[plane], *addrl) & bitmask ) | \
			( *addrl & ~bitmask));  \
	    } \
	    addrl += Poffset; \
	} \
	Addrl += nlwidth; \
	addrl = Addrl; \
    } \
    while (--len);

cfbPixVertS(rop, pixel, planemask, depth, addrl, nlwidth, x1, y1, len)
int rop;
unsigned long pixel;
unsigned long planemask;
int depth;
register int *addrl;	/* pointer to base of bitmap */
register int nlwidth;	/* width in longwords of bitmap */
int x1, y1;		/* initial point */
register int len;	/* length of line */
{
    register int plane;
    register int Poffset;
    int *Addrl;
    unsigned long bitmask = cfbmask[x1 & PIM];

    addrl = addrl + (y1 * nlwidth) + (x1 >> PWSH);

    if (len < 0)
    {
	nlwidth = -nlwidth;
	len = -len;
    }

    Poffset = cfbploff[1];
    Addrl = addrl;
    switch(rop)
    {
	case GXcopy:
	    VERT(fnCOPY);
	    break;
	case GXxor:
	    VERT(fnXOR);
	    break;
	case GXand:
	    VERT(fnAND);
	    break;
	case GXor:
	    VERT(fnOR);
	    break;
	case GXset:
	    VERT(fnSET);
	    break;
	case GXclear:
	    VERT(fnCLEAR);
	    break;
	case GXandReverse:
	    VERT(fnANDREVERSE);
	    break;
	case GXandInverted:
	    VERT(fnANDINVERTED);
	    break;
	case GXnor:
	    VERT(fnNOR);
	    break;
	case GXequiv:
	    VERT(fnEQUIV);
	    break;
	case GXinvert:
	    VERT(fnINVERT);
	    break;
	case GXorReverse:
	    VERT(fnORREVERSE);
	    break;
	case GXcopyInverted:
	    VERT(fnCOPYINVERTED);
	    break;
	case GXorInverted:
	    VERT(fnORINVERTED);
	    break;
	case GXnand:
	    VERT(fnNAND);
	    break;
	default:
	    break;
    }
}
