/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include "X.h"
#include "Xmd.h"
#include "servermd.h"

#include "misc.h"
#include "regionstr.h"
#include "gcstruct.h"
#include "windowstr.h"
#include "pixmapstr.h"
#include "scrnintstr.h"

#include "cfb.h"
#include "cfbmskbits.h"


/* cfbSetScanline -- copies the bits from psrc to the drawable starting at
 * (xStart, y) and continuing to (xEnd, y).  xOrigin tells us where psrc 
 * starts on the scanline. (I.e., if this scanline passes through multiple
 * boxes, we may not want to start grabbing bits at psrc but at some offset
 * further on.) 
 */
cfbWinSetScanline(y, xOrigin, xStart, xEnd, psrc, alu, pdstBase, widthDst, planemask, depth, srcjmp)
    int			y;
    int			xOrigin;	/* where this scanline starts */
    int			xStart;		/* first bit to use from scanline */
    int			xEnd;		/* last bit to use from scanline + 1 */
    register int	*psrc;
    int			alu;		/* raster op */
    int			*pdstBase;	/* start of the drawable */
    int			widthDst;	/* width of drawable in words */
    unsigned long	planemask;
    int			depth;
    int			srcjmp;
{
    int			w;		/* width of scanline in bits */
    register int	*pdst;		/* where to put the bits */
    int			dstBit;		/* offset in bits from beginning of 
					 * word */
    register int	nstart; 	/* number of bits from first partial */
    register int	nend; 		/* " " last partial word */
    int			offSrc;
    int		startmask, endmask, nlMiddle, nl;
    int			plane;
    int			dstjmp;

    pdst = pdstBase + (y * widthDst) + (xStart >> PWSH); 
    psrc += (xStart - xOrigin) >> PWSH;
    offSrc = (xStart - xOrigin) & PIM;
    w = xEnd - xStart;
    dstBit = xStart & PIM;
    dstjmp = getFbPlaneLongSize();

    for (plane = 0; plane < depth; plane++)
    {
	cfbplmk[plane] = (planemask >> plane) & 0x1;
    }

    if (alu == GXcopy)
    {

#ifdef FASTPUTBITS
	sx_set_operation_all(GXcopy, ALL_ONE);
#else
	sx_plane_select(planemask);
#endif

	if (dstBit + w <= PPW) 
	{ 
	    _TILEGANDPBITS(psrc, offSrc, dstBit, w, pdst, dstjmp, srcjmp, depth);
	} 
	else 
	{ 

	    maskbits(xStart, w, startmask, endmask, nlMiddle);
	    if (startmask) 
		nstart = PPW - dstBit; 
	    else 
		nstart = 0; 
	    if (endmask) 
		nend = xEnd & PIM; 
	    else 
		nend = 0; 
	    if (startmask) 
	    { 
		_TILEGANDPBITS(psrc, offSrc, dstBit, nstart, pdst, dstjmp, srcjmp, depth);
		pdst++; 
		offSrc += nstart;
		if (offSrc > PLST)
		{
		    psrc++;
		    offSrc -= PPW;
		}
	    } 
	    nl = nlMiddle; 
	    while (nl--) 
	    { 
		_TILEGANDPBITS0(psrc, offSrc, PPW, pdst, dstjmp, srcjmp, depth)
		pdst++; 
		psrc++; 
	    } 
	    if (endmask) 
	    { 
		_TILEGANDPBITS0(psrc, offSrc, nend, pdst, dstjmp, srcjmp, depth)
	    } 
	 
	} 
    }
    else
    {
	sx_plane_select(planemask);

	if (dstBit + w <= PPW) 
	{ 
	    _tilegandpbits(psrc, offSrc, dstBit, w, pdst, dstjmp, srcjmp, depth, alu);
	} 
	else 
	{ 

	    maskbits(xStart, w, startmask, endmask, nlMiddle);
	    if (startmask) 
		nstart = PPW - dstBit; 
	    else 
		nstart = 0; 
	    if (endmask) 
		nend = xEnd & PIM; 
	    else 
		nend = 0; 
	    if (startmask) 
	    { 
		_tilegandpbits(psrc, offSrc, dstBit, nstart, pdst,
						dstjmp, srcjmp, depth, alu);
		pdst++; 
		offSrc += nstart;
		if (offSrc > PLST)
		{
		    psrc++;
		    offSrc -= PPW;
		}
	    } 
	    nl = nlMiddle; 
	    while (nl--) 
	    { 
		_tilegandpbits0(psrc, offSrc, PPW, pdst, dstjmp, srcjmp, depth, alu)
		pdst++; 
		psrc++; 
	    } 
	    if (endmask) 
	    { 
		_tilegandpbits0(psrc, offSrc, nend, pdst, dstjmp, srcjmp, depth, alu)
	    } 
	 
	} 
    }
}

#define PSCANLINE(ROP) \
    if (dstBit + w <= PPW) \
    { \
	tilegandpbitsrop(psrc, offSrc, dstBit, w, pdst,\
				dstjmp, srcjmp, depth, ROP);\
    } \
    else \
    { \
	maskbits(xStart, w, startmask, endmask, nlMiddle);\
	if (startmask) \
	    nstart = PPW - dstBit; \
	else \
	    nstart = 0; \
	if (endmask) \
	    nend = xEnd & PIM; \
	else \
	    nend = 0; \
	if (startmask) \
	{ \
	    tilegandpbitsrop(psrc, offSrc, dstBit, nstart, pdst,\
				dstjmp, srcjmp, depth, ROP);\
	    pdst++; \
	    offSrc += nstart;\
	    if (offSrc > PLST)\
	    {\
		psrc++;\
		offSrc -= PPW;\
	    }\
	} \
	nl = nlMiddle; \
	while (nl--) \
	{ \
	    tilegandpbitsrop0(psrc, offSrc, PPW, pdst,\
				dstjmp, srcjmp, depth, ROP);\
	    pdst++; \
	    psrc++; \
	} \
	if (endmask) \
	{ \
	    tilegandpbitsrop0(psrc, offSrc, nend, pdst,\
				dstjmp, srcjmp, depth, ROP);\
	} \
    } 

cfbPixSetScanline(y, xOrigin, xStart, xEnd, psrc, alu, pdstBase, widthDst, planemask, depth, srcjmp, dstjmp)
    int			y;
    int			xOrigin;	/* where this scanline starts */
    int			xStart;		/* first bit to use from scanline */
    int			xEnd;		/* last bit to use from scanline + 1 */
    register int	*psrc;
    int			alu;		/* raster op */
    int			*pdstBase;	/* start of the drawable */
    int			widthDst;	/* width of drawable in words */
    unsigned long	planemask;
    int			depth;
    int			srcjmp;
    int			dstjmp;
{
    int			w;		/* width of scanline in bits */
    register int	*pdst;		/* where to put the bits */
    int			dstBit;		/* offset in bits from beginning of 
					 * word */
    register int	nstart; 	/* number of bits from first partial */
    register int	nend; 		/* " " last partial word */
    int			offSrc;
    int		startmask, endmask, nlMiddle, nl;
    int			plane;

    pdst = pdstBase + (y * widthDst) + (xStart >> PWSH); 
    psrc += (xStart - xOrigin) >> PWSH;
    offSrc = (xStart - xOrigin) & PIM;
    w = xEnd - xStart;
    dstBit = xStart & PIM;

    for (plane = 0; plane < depth; plane++)
    {
	cfbplmk[plane] = (planemask >> plane) & 0x1;
    }

    switch(alu)
    {
	case GXcopy:
	    PSCANLINE(fnCOPY);
	    break;
	case GXxor:
	    PSCANLINE(fnXOR);
	    break;
	case GXand:
	    PSCANLINE(fnAND);
	    break;
	case GXor:
	    PSCANLINE(fnOR);
	    break;
	case GXset:
	    PSCANLINE(fnSET);
	    break;
	case GXclear:
	    PSCANLINE(fnCLEAR);
	    break;
	case GXandReverse:
	    PSCANLINE(fnANDREVERSE);
	    break;
	case GXandInverted:
	    PSCANLINE(fnANDINVERTED);
	    break;
	case GXnor:
	    PSCANLINE(fnNOR);
	    break;
	case GXequiv:
	    PSCANLINE(fnEQUIV);
	    break;
	case GXinvert:
	    PSCANLINE(fnINVERT);
	    break;
	case GXorReverse:
	    PSCANLINE(fnORREVERSE);
	    break;
	case GXcopyInverted:
	    PSCANLINE(fnCOPYINVERTED);
	    break;
	case GXorInverted:
	    PSCANLINE(fnORINVERTED);
	    break;
	case GXnand:
	    PSCANLINE(fnNAND);
	    break;
	default:
	    break;
    }
}



/* SetSpans -- for each span copy pwidth[i] bits from psrc to pDrawable at
 * ppt[i] using the raster op from the GC.  If fSorted is TRUE, the scanlines
 * are in increasing Y order.
 * Source bit lines are server scanline padded so that they always begin
 * on a word boundary.
 */ 
void
cfbSetSpans(pDrawable, pGC, psrc, wMax, ppt, pwidth, nspans, fSorted)
    DrawablePtr		pDrawable;
    GCPtr		pGC;
    int			*psrc;
    int			wMax;
    register DDXPointPtr ppt;
    int			*pwidth;
    int			nspans;
    int			fSorted;
{
    int 		*pdstBase;	/* start of dst bitmap */
    int 		widthDst;	/* width of bitmap in words */
    register BoxPtr 	pbox, pboxLast, pboxTest;
    register DDXPointPtr pptLast;
    int 		alu;
    RegionPtr 		prgnDst;
    int			xStart, xEnd;
    int			yMax;
    int			type;
    int			srcpljmp, dstpljmp;
    OmronFbInfoPtr fbInfo;

    if (pGC->alu == GXnoop || pGC->planemask == 0)
	return;

    alu = pGC->alu;
    type = ((pDrawable->type == DRAWABLE_WINDOW) ? 1 : 0);
    prgnDst = ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip;

    pptLast = ppt + nspans;

    fbInfo = getOmronFbInfo(pDrawable);

    if (type)
    {
	pdstBase = (int *)getPlaneAddres(fbInfo);
	widthDst = (int)
		   ((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind
		    >> 2;
    }
    else
    {

	pdstBase = (int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	widthDst = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
	dstpljmp = widthDst * pDrawable->height;
	if(isCommonPlane( pdstBase, fbInfo ))
	    pdstBase = (int *)getPlaneAddres(fbInfo);
    }
    yMax = (int) pDrawable->y + (int) pDrawable->height;
    srcpljmp = (nspans * PixmapBytePad(wMax, 1)) >> 2;

    pbox = REGION_RECTS(prgnDst);
    pboxLast = pbox + REGION_NUM_RECTS(prgnDst);

    if(fSorted)
    {
    /* scan lines sorted in ascending order. Because they are sorted, we
     * don't have to check each scanline against each clip box.  We can be
     * sure that this scanline only has to be clipped to boxes at or after the
     * beginning of this y-band 
     */
	pboxTest = pbox;
	while(ppt < pptLast)
	{
	    pbox = pboxTest;
	    if(ppt->y >= yMax)
		break;
	    while(pbox < pboxLast)
	    {
		if(pbox->y1 > ppt->y)
		{
		    /* scanline is before clip box */
		    break;
		}
		else if(pbox->y2 <= ppt->y)
		{
		    /* clip box is before scanline */
		    pboxTest = ++pbox;
		    continue;
		}
		else if(pbox->x1 > ppt->x + *pwidth) 
		{
		    /* clip box is to right of scanline */
		    break;
		}
		else if(pbox->x2 <= ppt->x)
		{
		    /* scanline is to right of clip box */
		    pbox++;
		    continue;
		}

		/* at least some of the scanline is in the current clip box */
		xStart = max(pbox->x1, ppt->x);
		xEnd = min(ppt->x + *pwidth, pbox->x2);
		if (type)
		    cfbWinSetScanline(ppt->y, ppt->x, xStart, xEnd, psrc, alu,
			pdstBase, widthDst, pGC->planemask,
			pDrawable->depth, srcpljmp);
		else
		    cfbPixSetScanline(ppt->y, ppt->x, xStart, xEnd, psrc, alu,
			pdstBase, widthDst, pGC->planemask,
			pDrawable->depth, srcpljmp, dstpljmp);
		if(ppt->x + *pwidth <= pbox->x2)
		{
		    /* End of the line, as it were */
		    break;
		}
		else
		    pbox++;
	    }
	    /* We've tried this line against every box; it must be outside them
	     * all.  move on to the next point */
	    psrc += PixmapWidthInPadUnits(*pwidth, 1);
	    ppt++;
	    pwidth++;
	}
    }
    else
    {
    /* scan lines not sorted. We must clip each line against all the boxes */
	while(ppt < pptLast)
	{
	    if(ppt->y >= 0 && ppt->y < yMax)
	    {
		
		for(pbox = REGION_RECTS(prgnDst); pbox< pboxLast; pbox++)
		{
		    if(pbox->y1 > ppt->y)
		    {
			/* rest of clip region is above this scanline,
			 * skip it */
			break;
		    }
		    if(pbox->y2 <= ppt->y)
		    {
			/* clip box is below scanline */
			pbox++;
			break;
		    }
		    if(pbox->x1 <= ppt->x + *pwidth &&
		       pbox->x2 > ppt->x)
		    {
			xStart = max(pbox->x1, ppt->x);
			xEnd = min(pbox->x2, ppt->x + *pwidth);
			if (type)
			    cfbWinSetScanline(ppt->y, ppt->x, xStart, xEnd,
				psrc, alu, pdstBase, widthDst, pGC->planemask,
				pDrawable->depth, srcpljmp);
			else
			    cfbPixSetScanline(ppt->y, ppt->x, xStart, xEnd,
				psrc, alu, pdstBase, widthDst, pGC->planemask,
				pDrawable->depth, srcpljmp, dstpljmp);
		    }

		}
	    }
	    psrc += PixmapWidthInPadUnits(*pwidth, 1);
	    ppt++;
	    pwidth++;
	}
    }
}

