/***********************************************************
Copyright 1987 by Digital Equipment Corporation, Maynard, Massachusetts,
and the Massachusetts Institute of Technology, Cambridge, Massachusetts.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or MIT not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

******************************************************************/

/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include	"X.h"
#include	"Xmd.h"
#include	"Xproto.h"
#include	"fontstruct.h"
#include	"dixfontstr.h"
#include	"gcstruct.h"
#include	"windowstr.h"
#include	"scrnintstr.h"
#include	"pixmapstr.h"
#include	"regionstr.h"
#include	"cfb.h"
#include	"cfbmskbits.h"
#include	"cfbplanebit.h"

/*
    this works for fonts with glyphs <= 32 bits wide.

    This should be called only with a terminal-emulator font;
this means that the FIXED_METRICS flag is set, and that
glyphbounds == charbounds.

    in theory, this goes faster; even if it doesn't, it reduces the
flicker caused by writing a string over itself with image text (since
the background gets repainted per character instead of per string.)
this seems to be important for some converted X10 applications.

    Image text looks at the bits in the glyph and the fg and bg in the
GC.  it paints a rectangle, as defined in the protocol dcoument,
and the paints the characters.

   to avoid source proliferation, this file is compiled
two times:

*/

#if defined(FASTPUTBITS) && defined(FASTGETBITS)
#define FASTCHARS
#endif

/*
 * this macro "knows" that only characters <= 8 bits wide will
 * fit this case (which is why it is independent of GLYPHPADBYTES)
 */

#ifdef	RISC_CPU
#if (BITMAP_BIT_ORDER == MSBFirst) && (GLYPHPADBYTES != 4)
#if GLYPHPADBYTES == 1
#define ShiftAmnt   24
#else
#define ShiftAmnt   16
#endif

#define GetBits4    ((*char1++ << ShiftAmnt) | \
			SCRRIGHT (*char2++ << ShiftAmnt, xoff2) | \
			SCRRIGHT (*char3++ << ShiftAmnt, xoff3) | \
			SCRRIGHT (*char4++ << ShiftAmnt, xoff4))
#else
#define GetBits4    (*char1++ | \
			SCRRIGHT (*char2++, xoff2) | \
			SCRRIGHT (*char3++, xoff3) | \
			SCRRIGHT (*char4++, xoff4))
#endif
#else
#if (BITMAP_BIT_ORDER == MSBFirst) && (GLYPHPADBYTES != 4)
#if GLYPHPADBYTES == 1
#define ShiftAmnt   24
#else
#define ShiftAmnt   16
#endif

#define GetBits4    ((*char1++ << ShiftAmnt) | \
			SCRRIGHT (*char2++ << ShiftAmnt, xoff2))
#else
#define GetBits4    (*char1++ | SCRRIGHT (*char2++, xoff2))
#endif
#endif


#if GLYPHPADBYTES == 1
typedef	unsigned char	*glyphPointer;
#define USE_LEFTBITS
#endif

#if GLYPHPADBYTES == 2
typedef unsigned short	*glyphPointer;
#define USE_LEFTBITS
#endif

#if GLYPHPADBYTES == 4
typedef unsigned int	*glyphPointer;
#endif

#ifdef USE_LEFTBITS
#define GetBits1    getleftbits (char1, widthGlyph, c); \
		    c &= glyphMask; \
		    char1 = (glyphPointer) (((char *) char1) + glyphBytes);
#else
#define GetBits1    (*char1++)
#define	FONT_ZERO_PADDING
#endif

#ifdef	__GNUC__
#define	register
#endif

void
cfbTEGlyphBltMem(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase)
    DrawablePtr pDrawable;
    GC 		*pGC;
    int 	x, y;
    unsigned int nglyph;
    CharInfoPtr *ppci;		/* array of character info */
    unsigned char *pglyphBase;	/* start of array of glyphs */
{
    CharInfoPtr pci;
    FontInfoPtr pfi = pGC->font->pFI;
    int widthDst;
    unsigned int *pdstBase;	/* pointer to longword with top row 
				   of current glyph */

    int h;			/* height of glyph and char */
    int w;
    register int xpos;		/* current x  */
    int ypos;			/* current y */
    int widthGlyph;

    int hTmp;			/* counter for height */
    register int startmask, endmask;
    int nfirst;			/* used if glyphs spans a longword boundary */
    BoxRec bbox;		/* for clipping */
    int	widthGlyphs;
    register unsigned int  *dst;
	     unsigned int  *cls;
	     unsigned int  *dstsave;
    register unsigned int  c;
#ifdef	RISC_CPU
    register int	    xoff1, xoff2, xoff3, xoff4;
    register glyphPointer   char1, char2, char3, char4;
#else
    register int	    xoff1, xoff2;
    register glyphPointer   char1, char2;
#endif
    OmronFbInfoPtr fbInfo;
    int		   i;
    int 	   pljmp;

    unsigned long fg    = pGC->fgPixel;
    unsigned long bg    = pGC->bgPixel;
    unsigned long pmask = pGC->planemask;
    int           depth = pDrawable->depth;

#ifdef USE_LEFTBITS
    register int	    glyphMask;
    register unsigned int   tmpSrc;
    register int	    glyphBytes;
#endif

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	pdstBase = (unsigned int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	widthDst = (int)
		  (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
    }
    else
    {

	pdstBase = (unsigned int *)(((PixmapPtr)pDrawable)->devPrivate.ptr);
	widthDst = (int)(((PixmapPtr)pDrawable)->devKind) >> 2;
    }

    fbInfo = getOmronFbInfo( pDrawable );

    if(isCommonPlane(pdstBase,fbInfo))
    {
	pdstBase = (unsigned int *)getPlaneAddres(fbInfo);
	pljmp    =  getFbPlaneLongSize();
	sx_set_operation_all( GXcopy, ~0 );
    }
    else
    {
	pljmp    = widthDst * pDrawable->height;
    }

    xpos = x + pDrawable->x;
    ypos = y + pDrawable->y;

    pci = &pfi->maxbounds;
    widthGlyph = pci->metrics.characterWidth;
    h = pfi->fontAscent + pfi->fontDescent;
    w = widthGlyph * nglyph;

    xpos += pci->metrics.leftSideBearing;
    ypos -= pfi->fontAscent;

    bbox.x1 = xpos;
    bbox.x2 = xpos + w;
    bbox.y1 = ypos;
    bbox.y2 = ypos + h;

    switch ((*pGC->pScreen->RectIn)(
                ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip, &bbox))
    {
      case rgnPART:
	/* this is the WRONG thing to do, but it works.
	   calling the non-terminal text is easy, but slow, given
	   what we know about the font.

	   the right thing to do is something like:
	    for each clip rectangle
		compute at which row the glyph starts to be in it,
		   and at which row the glyph ceases to be in it
		compute which is the first glyph inside the left
		    edge, and the last one inside the right edge
		draw a fractional first glyph, using only
		    the rows we know are in
		draw all the whole glyphs, using the appropriate rows
		draw any pieces of the last glyph, using the right rows

	   this way, the code would take advantage of knowing that
	   all glyphs are the same height and don't overlap.

	   one day...
	*/
	cfbImageGlyphBlt(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase);
      case rgnOUT:
	return;
    }

    pdstBase += widthDst * ypos;

#ifdef	RISC_CPU
    widthGlyphs = widthGlyph << 2;
#else
    widthGlyphs = widthGlyph << 1;
#endif

    if( checkFgAndBg(fg, bg, pmask) )
    {
	cfbPlaneSetPixels(fg, bg, depth, pmask);
    }

    cls = pdstBase + (xpos >> 5);
    xoff1 = xpos & 0x1f;

    for(i = 0; i<depth; i++)
    {
	switch( cfbPlaneCopyType[i] )
	{
	case FillSet:
	    hTmp = h;
    	    if ((xoff1+w) <= 32)
    	    {
	        dst = cls;
		maskpartialbits(xoff1,w,startmask);
		while (hTmp--)
		{
		    *dst |= startmask;
	    	     dst += widthDst;
		}
    	    }
	    else
	    {
		int	nlm;
		register int nl;

		maskbits(xpos,w,startmask,endmask,nlm);

		dstsave = cls;
		while (hTmp--)
		{
		    dst = dstsave;
		    if(startmask)
		    {
			*dst++ |= startmask;
		    }
		    nl = nlm;
		    while(nl--)
		    {
			*dst++ = ~0;
		    }
		    if(endmask)
		    {
			*dst |= endmask;
		    }
		    dstsave += widthDst;
		}
	    }
	    break;

	case FillClear:
	    hTmp = h;
    	    if ((xoff1+w) <= 32)
    	    {
	        dst = cls;
		maskpartialbits(xoff1,w,startmask);
		while (hTmp--)
		{
		    *dst &= ~startmask;
	    	     dst += widthDst;
		}
    	    }
	    else
	    {
		int	nlm;
		register int nl;

		maskbits(xpos,w,startmask,endmask,nlm);

		dstsave = cls;

		while (hTmp--)
		{
		    dst = dstsave;
		    if(startmask)
		    {
			*dst++ &= ~startmask;
		    }
		    nl = nlm;
		    while(nl--)
		    {
			*dst++ = 0;
		    }
		    if(endmask)
		    {
			*dst &= ~endmask;
		    }
		    dstsave += widthDst;
		}
	    }
	    break;
        }
	cls += pljmp;
     }

#ifdef USE_LEFTBITS
    glyphMask = endtab[widthGlyph];
    glyphBytes = GLYPHWIDTHBYTESPADDED(pci);
#endif

#ifdef	RISC_CPU
    if (nglyph >= 4 && widthGlyphs <= 32)
    {
	while (nglyph >= 4)
	{
	    nglyph -= 4;
	    xoff1 = xpos & 0x1f;
	    xoff2 = widthGlyph;
	    xoff3 = xoff2 + widthGlyph;
	    xoff4 = xoff3 + widthGlyph;
	    char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char2 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char3 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char4 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
#else
    if (nglyph >= 2 && widthGlyphs <= 32)
    {
	while (nglyph >= 2)
	{
	    nglyph -= 2;
	    xoff1 = xpos & 0x1f;
	    xoff2 = widthGlyph;
	    char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char2 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
#endif

	    hTmp = h;
	    dstsave = pdstBase + (xpos >> 5);
#ifndef	FASTCHARS
	    if (xoff1 + widthGlyphs <= 32)
	    {
		maskpartialbits (xoff1, widthGlyphs, startmask);
#endif
		while (hTmp--)
		{
		    c = GetBits4;
#ifdef	FASTCHARS
#if BITMAP_BIT_ORDER == MSBFirst
		    c >>= 32 - widthGlyphs;
#endif
#endif
		    dst = dstsave;
		    for( i = 0; i<depth; i++)
		    {
			if(cfbPlaneCopyType[i] == CopySrc)
			{
#ifdef	FASTCHARS
			    FASTPUTBITS( c, xoff1, widthGlyphs,dst );
#else
			    *dst = (*dst & ~startmask) |
				   (SCRRIGHT(c, xoff1) & startmask);
#endif
			}
			else if(cfbPlaneCopyType[i] == CopyInv)
			{
#ifdef	FASTCHARS
			    FASTPUTBITS( ~c, xoff1, widthGlyphs,dst );
#else
			    *dst = (*dst & ~startmask) |
				   (SCRRIGHT(~c, xoff1) & startmask);
#endif
			}
			dst += pljmp;
		    }
		    dstsave += widthDst;
		}
#ifndef	FASTCHARS
	    }
	    else
	    {
		mask32bits (xoff1, widthGlyphs, startmask, endmask);
		nfirst = 32 - xoff1;
		while (hTmp--)
		{
		    c = GetBits4;
		    dst = dstsave;
		    for( i = 0; i<depth; i++)
		    {
			if(cfbPlaneCopyType[i] == CopySrc)
			{
			    *dst = (*dst & ~startmask) |
				   (SCRRIGHT(c,xoff1) & startmask);
			    *(dst+1) = (*(dst+1) & ~endmask) |
				   (SCRLEFT(c,nfirst) & endmask);
			}
			else if(cfbPlaneCopyType[i] == CopyInv)
			{
			    *dst = (*dst & ~startmask) |
				   (SCRRIGHT(~c,xoff1) & startmask);
			    *(dst+1) = (*(dst+1) & ~endmask) |
				   (SCRLEFT(~c,nfirst) & endmask);
			}
			dst += pljmp;
		    }
		    dstsave += widthDst;
		}
	    }
#endif
	    xpos += widthGlyphs;
	}
    }

    while(nglyph--)
    {
	xoff1 = xpos & 0x1f;
	char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	hTmp = h;
	dstsave = pdstBase + (xpos >> 5);

#ifndef	FASTCHARS
	if (xoff1 + widthGlyph <= 32)
	{
	    maskpartialbits (xoff1, widthGlyph, startmask);
#endif
	    while (hTmp--)
	    {
#ifdef FASTCHARS
#ifdef USE_LEFTBITS
		FASTGETBITS (char1,0,widthGlyph,c);
		char1 = (glyphPointer) (((char *) char1) + glyphBytes);
#else
		c = *char1++;
#if BITMAP_BIT_ORDER == MSBFirst
		c >>= 32 - widthGlyph;
#endif
#endif
#else
#ifdef USE_LEFTBITS
		GetBits1
#else
		c = GetBits1;
#endif
#endif
		dst = dstsave;
		for( i = 0; i<depth; i++)
		{
		    if(cfbPlaneCopyType[i] == CopySrc)
		    {
#ifdef	FASTCHARS
			FASTPUTBITS (c,xoff1,widthGlyph,dst);
#else
			*dst = (*dst & ~startmask) |
			       (SCRRIGHT(c, xoff1) & startmask);
#endif
		    }
		    else if(cfbPlaneCopyType[i] == CopyInv)
		    {
#ifdef	FASTCHARS
			FASTPUTBITS (~c,xoff1,widthGlyph,dst);
#else
			*dst = (*dst & ~startmask) |
			       (~SCRRIGHT(c, xoff1) & startmask);
#endif
		    }
		    dst += pljmp;
		}
		dstsave += widthDst;
	    }
#ifndef	FASTCHARS
	}
	else
	{
	    mask32bits (xoff1, widthGlyph, startmask, endmask);
	    nfirst = 32 - xoff1;
	    while (hTmp--)
	    {
#ifdef USE_LEFTBITS
		GetBits1
#else
		c = GetBits1;
#endif
		dst = dstsave;
		for( i = 0; i<depth; i++)
		{
		    if(cfbPlaneCopyType[i] == CopySrc)
		    {
			*dst = (*dst & ~startmask) |
			       (SCRRIGHT(c, xoff1) & startmask);
			*(dst+1) = (*(dst+1) & ~endmask) |
			           (SCRLEFT(c, nfirst) & endmask);
		    }
		    else if(cfbPlaneCopyType[i] == CopyInv)
		    {
			*dst = (*dst & ~startmask) |
			       (~SCRRIGHT(c, xoff1) & startmask);
			*(dst+1) = (*(dst+1) & ~endmask) |
			           (~SCRLEFT(c, nfirst) & endmask);
		    }
		    dst += pljmp;
		}
		dstsave += widthDst;
	    }
	}
#endif
	xpos += widthGlyph;
    }
}


void
cfbTEGlyphBlt(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase)
    DrawablePtr pDrawable;
    GC 		*pGC;
    int 	x, y;
    unsigned int nglyph;
    CharInfoPtr *ppci;		/* array of character info */
    unsigned char *pglyphBase;	/* start of array of glyphs */
{
    CharInfoPtr pci;
    FontInfoPtr pfi = pGC->font->pFI;
    int widthDst;
    unsigned int *pdstBase;	/* pointer to longword with top row 
				   of current glyph */

    int h;			/* height of glyph and char */
    int w;
    register int xpos;		/* current x  */
    int ypos;			/* current y */
    int widthGlyph;

    int hTmp;			/* counter for height */
    register int startmask, endmask;
    int nfirst;			/* used if glyphs spans a longword boundary */
    BoxRec bbox;		/* for clipping */
    int	widthGlyphs;
    register unsigned int  *dst;
    register unsigned int  c;
#ifdef	RISC_CPU
    register int	    xoff1, xoff2, xoff3, xoff4;
    register glyphPointer   char1, char2, char3, char4;
#else
    register int	    xoff1, xoff2;
    register glyphPointer   char1, char2;
#endif

    unsigned long fg    = pGC->fgPixel;
    unsigned long bg    = pGC->bgPixel;
    unsigned long pmask = pGC->planemask;
    int           depth = pDrawable->depth;

#ifdef USE_LEFTBITS
    register int	    glyphMask;
    register unsigned int  tmpSrc;
    register int	    glyphBytes;
#endif

    if (pDrawable->type == DRAWABLE_WINDOW)
    {
	pdstBase = (unsigned int *)
		(((PixmapPtr)(pDrawable->pScreen->devPrivate))->devPrivate.ptr);
	widthDst = (int)
		  (((PixmapPtr)(pDrawable->pScreen->devPrivate))->devKind) >> 2;
    }
    else
    {
	cfbTEGlyphBltMem(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase);
	return;
    }

    xpos = x + pDrawable->x;
    ypos = y + pDrawable->y;

    pci = &pfi->maxbounds;
    widthGlyph = pci->metrics.characterWidth;
    h = pfi->fontAscent + pfi->fontDescent;
    w = widthGlyph * nglyph;

    xpos += pci->metrics.leftSideBearing;
    ypos -= pfi->fontAscent;

    bbox.x1 = xpos;
    bbox.x2 = xpos + w;
    bbox.y1 = ypos;
    bbox.y2 = ypos + h;

    switch ((*pGC->pScreen->RectIn)(
                ((cfbPrivGC *)(pGC->devPrivates[cfbGCPrivateIndex].ptr))->pCompositeClip, &bbox))
    {
      case rgnPART:
	/* this is the WRONG thing to do, but it works.
	   calling the non-terminal text is easy, but slow, given
	   what we know about the font.

	   the right thing to do is something like:
	    for each clip rectangle
		compute at which row the glyph starts to be in it,
		   and at which row the glyph ceases to be in it
		compute which is the first glyph inside the left
		    edge, and the last one inside the right edge
		draw a fractional first glyph, using only
		    the rows we know are in
		draw all the whole glyphs, using the appropriate rows
		draw any pieces of the last glyph, using the right rows

	   this way, the code would take advantage of knowing that
	   all glyphs are the same height and don't overlap.

	   one day...
	*/
	cfbImageGlyphBlt(pDrawable, pGC, x, y, nglyph, ppci, pglyphBase);
      case rgnOUT:
	return;
    }

    pdstBase += widthDst * ypos;
#ifdef	RISC_CPU
    widthGlyphs = widthGlyph << 2;
#else
    widthGlyphs = widthGlyph << 1;
#endif

    sx_set_transparent(GXcopy, ~0, bg);
    sx_plane_select(pmask);

    hTmp = h;
    dst = pdstBase + (xpos >> 5);
    xoff1 = xpos & 0x1f;
    if((w + xoff1) <= 32)
    {
	maskpartialbits (xoff1, w, startmask);
	while(hTmp--)
	{
	    *dst = startmask;
	    dst  += widthDst;
	}
    }
    else
    {
	register int nl;
        register unsigned int *cls;
	int nlm;

	maskbits (xoff1, w, startmask, endmask, nlm);

	while(hTmp--)
	{
	    cls = dst;
	    if(startmask)
	    {
		*cls++ = startmask;
	    }
	    nl = nlm;
	    while(nl--)
	    {
		*cls++ = ~0;
	    }
	    if(endmask)
	    {
		*cls = endmask;
	    }
	    dst  += widthDst;
	}
    }

    sx_set_transparent(GXcopy, ~0, fg);
    sx_plane_select(pmask);

#ifdef USE_LEFTBITS
    glyphMask = endtab[widthGlyph];
    glyphBytes = GLYPHWIDTHBYTESPADDED(pci);
#endif

#ifdef	RISC_CPU
    if (nglyph >= 4 && widthGlyphs <= 32)
    {
	while (nglyph >= 4)
	{
	    nglyph -= 4;
	    xoff1 = xpos & 0x1f;
	    xoff2 = widthGlyph;
	    xoff3 = xoff2 + widthGlyph;
	    xoff4 = xoff3 + widthGlyph;
	    char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char2 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char3 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char4 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
#else
    if (nglyph >= 2 && widthGlyphs <= 32)
    {
	while (nglyph >= 2)
	{
	    nglyph -= 2;
	    xoff1 = xpos & 0x1f;
	    xoff2 = widthGlyph;
	    char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	    char2 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
#endif

	    hTmp = h;
	    dst = pdstBase + (xpos >> 5);

	    if (xoff1 + widthGlyphs <= 32)
	    {
#ifndef	FONT_ZERO_PADDING
		maskpartialbits (xoff1, widthGlyphs, startmask);
#endif
		while (hTmp--)
		{
#ifdef	FONT_ZERO_PADDING
		    *(dst) = SCRRIGHT(GetBits4, xoff1);
#else
		    *(dst) = SCRRIGHT(GetBits4, xoff1) & startmask;
#endif
		    dst += widthDst;
		}
	    }
	    else
	    {
#ifndef	FONT_ZERO_PADDING
		mask32bits (xoff1, widthGlyphs, startmask, endmask);
#endif
		nfirst = 32 - xoff1;
		while (hTmp--)
		{
		    c = GetBits4;
#ifdef	FONT_ZERO_PADDING
		    *(dst) = SCRRIGHT(c,xoff1);
		    *(dst+1) = SCRLEFT(c,nfirst);
#else
		    *dst = SCRRIGHT(c,xoff1) & startmask;
		    *(dst+1) = SCRLEFT(c,nfirst) & endmask;
#endif
		    dst += widthDst;
		}
	    }
	    xpos += widthGlyphs;
	}
    }

    while(nglyph--)
    {
	xoff1 = xpos & 0x1f;
	char1 = (glyphPointer) (pglyphBase + (*ppci++)->byteOffset);
	hTmp = h;
	dst = pdstBase + (xpos >> 5);

	if (xoff1 + widthGlyph <= 32)
	{
#ifndef	FONT_ZERO_PADDING
	    maskpartialbits (xoff1, widthGlyph, startmask);
#endif
	    while (hTmp--)
	    {
#ifdef USE_LEFTBITS
		GetBits1
		(*dst) = SCRRIGHT(c, xoff1) & startmask;
#else
#ifdef	FONT_ZERO_PADDING
		(*dst) = SCRRIGHT(GetBits1, xoff1);
#else
		(*dst) = SCRRIGHT(GetBits1, xoff1) & startmask;
#endif
#endif
		dst += widthDst;
	    }
	}
	else
	{
#ifndef	FONT_ZERO_PADDING
	    mask32bits (xoff1, widthGlyph, startmask, endmask);
#endif
	    nfirst = 32 - xoff1;
	    while (hTmp--)
	    {
#ifdef USE_LEFTBITS
		GetBits1
		*dst = SCRRIGHT(c,xoff1) & startmask;
		*(dst+1) = SCRLEFT(c,nfirst) & endmask;
#else
		c = GetBits1;
#ifdef	FONT_ZERO_PADDING
		*dst = SCRRIGHT(c,xoff1);
		*(dst+1) = SCRLEFT(c,nfirst);
#else
		*dst = SCRRIGHT(c,xoff1) & startmask;
		*(dst+1) = SCRLEFT(c,nfirst) & endmask;
#endif
#endif
		dst += widthDst;
	    }
	}
	xpos += widthGlyph;
    }
}
