/******************************************************************
(c) Copyright OMRON Corporation 1990

                        All Rights Reserved

******************************************************************/

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#ifdef	UniOSU
#include <termio.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/kbms.h>
#include <sys/deftbl.t>
#include <sys/semienc.t>
#include <sys/param.h>
#else
#ifdef NO_SIGIO
#include <sys/param.h>
#endif
#include <om68kdev/kbms.h>
#include <om68kdev/deftbl.t>
#include <om68kdev/keymap.t>
#endif

#ifndef UniOSU 
#define NOTREP 0
#define REPEAT 1
#endif


#include "X.h"
#include "Xmd.h"
#define NEED_EVENTS
#include "Xproto.h"

#include "scrnintstr.h"

#include "input.h"

#include "omronKbd.h"
#if	defined(UniOSU) || defined(NO_SIGIO)
#include "osdep.h"
#endif

extern CARD8 *omronKeyModMap[];
extern KeySymsRec omronKeySyms[];


static void omronBell();
static void omronKbdCtrl();


extern int lastEventTime;

int keyboardfd = -1;
int omron_key_state = 0;
#if  defined(UniOSU) || defined(NO_SIGIO)
long omronKeyboardMask[mskcnt];
#endif
#ifdef UniOSU
static struct termio old_term;
#else
static struct sgttyb old_term;
#endif

static KeybdCtrl omronKeybdCtrl;
static unsigned char omronAutoRepeats[32] = {
	0,    0xff, 0x81, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xc0, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x3f
};

char omronSemiEncodeDef[CODTBSZ];
char omronSemiEncode[CODTBSZ];
char omronSemiKanaEncode[CODTBSZ];

static unsigned char kana_offset; 
static KeyCode minkey,maxkey; 
static KeyCode kana_minkey,kana_maxkey; 
  

int
omronKbdProc(pKeyboard,what)
DevicePtr     pKeyboard;
int           what;
{
	switch(what) {
		case DEVICE_INIT:
			omronKbdInit();
#if	defined(UniOSU) || defined(NO_SIGIO)
			CLEARBITS(omronKeyboardMask);
#endif
			bcopy (omronAutoRepeats, defaultKeyboardControl.autoRepeats,
			sizeof (omronAutoRepeats));
			omronKeybdCtrl = defaultKeyboardControl;
			kana_offset = (omronKeySyms[0].maxKeyCode + omronKeySyms[0].minKeyCode
					 + 1)/2 - omronKeySyms[0].minKeyCode ; 
			pKeyboard->on = FALSE;
			InitKeyboardDeviceStruct(pKeyboard,
				&(omronKeySyms[0]),
				(omronKeyModMap[0]),
				omronBell,
				omronKbdCtrl);
			break;
		case DEVICE_ON:
			AddEnabledDevice(keyboardfd);
#if	defined(UniOSU) || defined(NO_SIGIO)
			BITSET(omronKeyboardMask,keyboardfd);
#endif
			pKeyboard->on = TRUE;
			break;
		case DEVICE_CLOSE:
		case DEVICE_OFF:
			RemoveEnabledDevice(keyboardfd);
#if	defined(UniOSU) || defined(NO_SIGIO)
			BITCLEAR(omronKeyboardMask,keyboardfd);
#endif
			omronKbdReset();
			pKeyboard->on = FALSE;
			break;
	}
	return (Success);
} 


omronKbdInit()
{
#ifdef UniOSU
	struct termio new_term;
#else
	struct sgttyb new_term;
	register int i;
#endif
	struct kbssgeta segeta;


	if((keyboardfd = open("/dev/kbd",O_RDWR,0)) < 0) {
		Error("Can't open /dev/kbd.");
	}

#if	defined(UniOSU) || defined(NO_SIGIO)
	{
		int arg = 1; 
		ioctl (keyboardfd, FIONBIO, &arg);
	}
#else
	if (fcntl(keyboardfd, F_SETFL, (FNDELAY|FASYNC)) < 0
	    || fcntl(keyboardfd, F_SETOWN, getpid()) < 0) {
		ErrorF("Can't set up keyboard on fd %d.",keyboardfd);
	}
#endif	/* UniOSU */

	if(ioctl(keyboardfd,KBSETM,SEMIENCODED) < 0) {
		Error("keyboard ioctl SEMIENCODED fault.");
	}

	if (ioctl(keyboardfd,KBRSTENABLE,RSTDISABLE) < 0) {
		Error("keyboard ioctl RSTDISABLE fault.");
	}
#ifdef UniOSU
	if(ioctl(keyboardfd,TCGETA,&old_term) < 0) {
		Error("keyboard ioctl TCGETA fault.");
	}
#else
	if(ioctl(keyboardfd,TIOCGETP,&old_term) < 0) {
		Error("keyboard ioctl TCGETA fault.");
	}
#endif

	new_term = old_term;

#ifdef UniOSU
	new_term.c_lflag &= ~ICANON;
	new_term.c_lflag &= ~ISIG;
	new_term.c_cc[VEOF] = 1;
	if (ioctl(keyboardfd,TCSETA,&new_term) < 0) {	
		Error("keyboard ioctl TCSETA fault.");
	}
#else
	new_term.sg_flags = RAW;
	if (ioctl(keyboardfd,TIOCSETP,&new_term) < 0) {	
		Error("keyboard ioctl TIOCSETP fault.");
	}
#endif
	
	segeta.kbsreptp = omronSemiEncodeDef;
	if(ioctl(keyboardfd, KBSGETTA, &segeta) < 0) {
		Error("kbd ioctl KBSGETTA fault.");
	}
#ifndef UniOSU
	for(i = 0 ; i < CODTBSZ ; i++ ) {
		omronSemiEncodeDef[i] = omronSemiEncodeDef[i]?0:1;
	}
#endif
	bcopy(omronSemiEncodeDef,omronSemiEncode,CODTBSZ);
	bcopy(omronSemiEncodeDef,omronSemiKanaEncode,CODTBSZ);
}
		
omronKbdReset()
{
	/*
	rep.repstart = REPCNT;
	rep.repinterval = RPINTVL;
	if(ioctl(keyboardfd, KBREPTIM, &rep) < 0) {
                 Error("kbd ioctl fault(reset repeat mode).");
        }

	if(ioctl(keyboardfd, KBCLICK, 0) < 0) {
		Error("kbd ioctl fault(reset click).");
	}
	*/

	if (ioctl(keyboardfd,KBSETM,ENCODED) < 0) {
		Error("kbd ioctl fault(set encode mode).");
	}

	if (ioctl(keyboardfd,KBRSTENABLE,RSTENABLE) < 0) {
		Error("Kbd ioctl fault.\n");
	}

#ifdef UniOSU	
	if (ioctl(keyboardfd,TCSETA,&old_term) < 0) {
		Error("kbd ioctl fault(reset term mode).");
	}
#else
	if (ioctl(keyboardfd,TIOCSETP,&old_term) < 0) {
		Error("kbd ioctl fault(reset term mode).");
	}
#endif

	close(keyboardfd);

}
		
omronProcessKbdEvent(pKeyboard)
DevicePtr	pKeyboard;
{
	int n;
	register int i;

	if (ioctl(keyboardfd,FIONREAD,&n) == -1) {
		Error("ioctl FIONREAD error");
		return FALSE;
	}

	for ( i = 0 ; i < n ; i++ )
		omronKbdSetEvent(pKeyboard);

	return n;
}

omronKbdSetEvent(pKeyboard)
DevicePtr	pKeyboard;
{

	register int i;
	int	key_state = 0;
	unsigned char c;
	xEvent xE;
	struct kbssgeta seseta;

	read(keyboardfd,&c,1);

	lastEventTime = GetTimeInMillis();
	xE.u.keyButtonPointer.time = lastEventTime;

	miPointerPosition (screenInfo.screens[0],
		&xE.u.keyButtonPointer.rootX,
		&xE.u.keyButtonPointer.rootY);

	if (!(c & 0x80)) { /* key down */
		c &= 0x7f;
		switch (c) {
		case CAPSLOCK_KEY:
			key_state = KS_CAPSLOCK;
			break;
		case LEFT_SHIFT_KEY:
		case RIGHT_SHIFT_KEY:
			key_state = KS_SHIFT;
			break;
		case CTRL_KEY:
			key_state = KS_CTRL;
			break;
		case SPECIAL_KEY:
			key_state = KS_SPECIAL;
			break;
		case KANA_KEY:
			key_state = KS_KANA;
			omron_key_state |= key_state;
			seseta.kbsreptp = omronSemiKanaEncode;
			if(ioctl(keyboardfd, KBSSETTA, &seseta) < 0) {
				Error("kbd ioctl fault(semi-encode table set)");
			}
			return;
		default:
			key_state = 0;
		}
		omron_key_state |= key_state;
		xE.u.u.type = KeyPress;
	} else {
		c &= 0x7f;
		switch (c) {
		case CAPSLOCK_KEY:
			key_state = KS_CAPSLOCK;
			break;
		case LEFT_SHIFT_KEY:
		case RIGHT_SHIFT_KEY:
			key_state = KS_SHIFT;
			break;
		case CTRL_KEY:
			key_state = KS_CTRL;
			break;
		case SPECIAL_KEY:
			key_state = KS_SPECIAL;
			break;
		case KANA_KEY:
			key_state = KS_KANA;
			omron_key_state &= ~key_state;
			seseta.kbsreptp = omronSemiEncode;
			if(ioctl(keyboardfd, KBSSETTA, &seseta) < 0) {
				Error("kbd ioctl fault(semi-encode table set)");
			}
			return;
		default:
			key_state = 0;
		}
		omron_key_state &= ~key_state;
		xE.u.u.type = KeyRelease;
	}
#define	EXIT_KEY	0x34
	if ( ( c == EXIT_KEY ) && ( omron_key_state == (KS_SHIFT | KS_CTRL | KS_SPECIAL) ) ) {
		GiveUp();
	}
	if ( (omron_key_state & KS_KANA) && (!(omron_key_state & KS_CTRL)) && (c != CTRL_KEY)
		&& (c != LEFT_SHIFT_KEY) && (c != RIGHT_SHIFT_KEY) && (c != CAPSLOCK_KEY) && (c != SPECIAL_KEY) )  {
		xE.u.u.detail = c + kana_offset;
	} else {
		xE.u.u.detail = c;
	}

	(* pKeyboard->processInputProc) (&xE, pKeyboard, 1);
}
	
static void
omronBell(loudness, pKeyboard)
int           loudness;
DevicePtr     pKeyboard;
{
	register int 	i,hz,msec;
	char c;

	if (loudness == 0) {
		return;
	}
	hz = omronKeybdCtrl.bell_pitch;
#ifdef UniOSU
	if (hz <= 100){
		c = 0x47;
	} else  if (hz <= 150) {
		c = 0x46;
	} else  if (hz <= 300) {
		c = 0x45;
	} else  if (hz <= 600) {
		c = 0x44;
	} else  if (hz <= 1000) {
		c = 0x43;
	} else  if (hz <= 1500) {
		c = 0x42;
	} else  if (hz <= 3000) {
		c = 0x41;
	} else {
		c = 0x40;
	}

	msec = omronKeybdCtrl.bell_duration;
	if (msec <= 40) {
		i = 0;
	} else if (msec <= 150) {
		i = 1;
	} else if (msec <= 400) {
		i = 2;
	} else {
		i = 3;
	}
	c |= (i << 3);
#else
	if (hz <= 100){
		c = K_HZ100;
	} else  if (hz <= 150) {
		c = K_HZ150;
	} else  if (hz <= 300) {
		c = K_HZ300;
	} else  if (hz <= 600) {
		c = K_HZ600;
	} else  if (hz <= 1000) {
		c = K_HZ1000;
	} else  if (hz <= 1500) {
		c = K_HZ1500;
	} else  if (hz <= 3000) {
		c = K_HZ3000;
	} else {
		c = K_HZ6000;
	}

	msec = omronKeybdCtrl.bell_duration;
	if (msec <= 40) {
		c |= K_TIME40;
	} else if (msec <= 150) {
		c |= K_TIME150;
	} else if (msec <= 400) {
		c |= K_TIME400;
	} else {
		c |= K_TIME700;
	}
#endif
#ifdef UniOSU
	if (  write(keyboardfd,&c,1) < 0 ) {
		Error ("omronBell can't write keyboard.");
	}
#else
	if (  ioctl(keyboardfd, KBBELL, c) < 0 ) {
		Error ("omronBell ioctl KBBELL error.");
	}
#endif
}

static void
omronKbdCtrl(pKeyboard, ctrl)
DevicePtr     pKeyboard;
KeybdCtrl     *ctrl;
{
	struct kbrepeat rep;
	struct kbssgeta seseta;
	register KeyCode key;
	register int mask,i;
	int shift;

	if ( pKeyboard->on == FALSE )   	/* init */
	{
		if ( ctrl->click == 0 )
		{
			if(ioctl(keyboardfd, KBCLICK, 0) < 0) {
				Error("kbd ioctl KBCLICK error");
			}
		} else {
			if(ioctl(keyboardfd, KBCLICK, 5) < 0) {
				Error("kbd ioctl KBCLICK error");
			}
		}

		if ( ctrl->autoRepeat == TRUE )
			rep.repstart = REPCNT;
		else
			rep.repstart = 0;
		
		rep.repinterval = RPINTVL;
	
		if(ioctl(keyboardfd, KBREPTIM, &rep) < 0) {
			Error("kbd ioctl KBREPTIM error");
		} 

		minkey = omronKeySyms[0].minKeyCode;
		maxkey = omronKeySyms[0].minKeyCode + kana_offset -1; 

		for( key = minkey ; key <= maxkey ; key++)
		{
			if ( omronSemiEncodeDef[key] == REPEAT )
			{
				i = (key >> 3) ; 
				mask = (1 << (key & 7));
				if ( ctrl->autoRepeats[i] & mask ) {
					omronSemiEncode[key] = REPEAT;	
				} else {
					omronSemiEncode[key] = NOTREP;	
				}
			}
		}

		kana_minkey = omronKeySyms[0].minKeyCode + kana_offset;
		kana_maxkey = omronKeySyms[0].maxKeyCode; 

		for( key = kana_minkey ; key <= kana_maxkey ; key++)
		{
			if ( omronSemiEncodeDef[key - kana_offset] == REPEAT )
			{
				i = (key >> 3) ; 
				mask = (1 << (key & 7));
				if ( ctrl->autoRepeats[i] & mask ) {
					omronSemiKanaEncode[key - kana_offset] = REPEAT;	
				} else {
					omronSemiKanaEncode[key - kana_offset] = NOTREP;	
				}
			}
		}

		seseta.kbsreptp = omronSemiEncode;
		if(ioctl(keyboardfd, KBSSETTA, &seseta) < 0) {
			Error("kbd ioctl KBSSETTA error");
		} 
	} else {
		if ( ctrl->click != omronKeybdCtrl.click )
		{
			if( ctrl->click == 0 ) {
				if(ioctl(keyboardfd, KBCLICK, 0) < 0) {
					Error("kbd ioctl KBCLICK error");
				}
			} else {
				if(ioctl(keyboardfd, KBCLICK, 5) < 0) {
					Error("kbd ioctl KBCLICK error");
				}
			}
		}
		if ( ctrl->autoRepeat != omronKeybdCtrl.autoRepeat )
		{ 
			if ( ctrl->autoRepeat == TRUE )	 
				rep.repstart = REPCNT;
			else
				rep.repstart = 0;
			rep.repinterval = RPINTVL;
			if(ioctl(keyboardfd, KBREPTIM, &rep) < 0) {
				Error("kbd ioctl KBREPTIM error");
			}
		}
		for ( i = 0 ; i < 32 ; i++) {
			if ( mask = ctrl-> autoRepeats[i] ^ omronKeybdCtrl.autoRepeats[i] )
			{	
				shift = ffs(mask) -1;
				key = ( i << 3) | shift;
				if ( key >= kana_minkey && key <= kana_maxkey )
				{
					key -= kana_offset;
					if ( omronSemiEncodeDef[key] == REPEAT ) {
						if ( ctrl->autoRepeats[i] & (1 << shift) )
							omronSemiKanaEncode[key] = REPEAT;
						else
							omronSemiKanaEncode[key] = NOTREP;
						if ( omron_key_state & KS_KANA ) {
							seseta.kbsreptp = omronSemiKanaEncode;
							if(ioctl(keyboardfd, KBSSETTA, &seseta) < 0) {
								Error("kbd ioctl fault(semi-encode table set)");
							}
						}
					}
				}
				else if ( key >= minkey && key <= maxkey ) {
					if ( omronSemiEncodeDef[key] == REPEAT ) {
						if ( ctrl->autoRepeats[i] & (1 << shift) )
						{
							omronSemiEncode[key] = REPEAT;
						}else{
							omronSemiEncode[key] = NOTREP;
						}
						if ( !(omron_key_state & KS_KANA) ){
							seseta.kbsreptp = omronSemiEncode;
							if(ioctl(keyboardfd, KBSSETTA, &seseta) < 0) {
								Error("kbd ioctl fault(semi-encode table set)");
							}
						}
					}
				}
				break;
			}
		}
	}
	omronKeybdCtrl = *ctrl;
}
	
Bool
LegalModifier(key)
BYTE key;
{
	return(TRUE);
}
		

